(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
This is the Addendum.m file, a Mathematica routine for FeynRules that creates external files which
will allow us to establish an interface between the FeynRules environment, on the one hand, and both
QGRAF and Amperate, on the other. For QGRAF, we generate the model file: the file which shall be read
by QGRAF as describing the model. As for Amperate, we generate several files with the Feynman rules 
of the model.

Created by: Duarte Fontes
Email: duarte.fontes@kit.edu
Last update: 2/April/2025
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

(* Get the Feynman rules for the vertices*)
If[ToString[InputForm[LGauge]] != "LGauge", vGauge = FeynmanRules[LGauge, MaxParticles -> 6]];
If[ToString[InputForm[LHiggs]] != "LHiggs", vHiggs = FeynmanRules[LHiggs, MaxParticles -> 6]];
If[ToString[InputForm[LGhost]] != "LGhost", vGhosts = FeynmanRules[LGhost, MaxParticles -> 6]];
If[ToString[InputForm[LFermions]] != "LFermions", vFermions = FeynmanRules[WeylToDirac[LFermions], FlavorExpand -> True, MaxParticles -> 6]];
If[ToString[InputForm[LYukawa]] != "LYukawa", vYukawa = FeynmanRules[WeylToDirac[LYukawa], FlavorExpand -> True, MaxParticles -> 6]];
If[ToString[InputForm[LBeaks]] != "LBeaks", vBeaks = FeynmanRules[LBeaks, FlavorExpand -> True]];

propsrulestot={};
propsrulesexp={};
If[PrMassFL,
(* Determine the rules for the propagators in case the masses should be extracted from the Lagrangean *)
(* N.B.: the fermionic propagators are never determined in this way *)
LProppre = 0;
If[ToString[InputForm[LGauge]] != "LGauge", LProppre += ExpandIndices[LGauge]];
If[ToString[InputForm[LHiggs]] != "LHiggs", LProppre += ExpandIndices[LHiggs]];
If[ToString[InputForm[LGhost]] != "LGhost", LProppre += ExpandIndices[LGhost]];
If[ToString[InputForm[LGF]] != "LGF", LProppre += ExpandIndices[LGF]];
LProppre = Expand[LProppre];
LProp=ExpandIndices[LProppre, MaxParticles -> 2];
LP1 = LProp /. insetas;
LP2 = LP1 /. consrules;
LP3 = (LP2 /. stilletas) // Expand;
LP4a = LP3 /. consrules;
LP4b = LP4a /. repsetas /. repsetas /. repsetas /. repsetas;
LP5 = (LP4b /. proprep);
Do[
LP6a = (((Coefficient[LP5, propsetas[[i]]] // Simplify) /. MyRestr) // FullSimplify) /. MyRestr;
LP6b = ((LP6a // Expand) // FullSimplify) /. MyRestr;
LP6c = FullSimplify[LP6b/I];
If[ToString[InputForm[propstypes[[i]]]]=="S"||ToString[InputForm[propstypes[[i]]]]=="U",
LP7 = 1/LP6c;
LP8 = Parc[LP7 // Expand];
If[LP7 == LP8,
	LP9 = Denominator[LP8];
	If[Coefficient[LP9,p,2]==-1,propmass=mysqrt[LP9/.p->0],propmass=mysqrt[-LP9/.p->0]];
	If[colcoin!=0&&ToString[InputForm[propsnames[[i]]]]==ToString[InputForm[myglugh]],
		propsrulestot=Append[propsrulestot,"prop" <> propsnames[[i]] <> "[m_,p_,J1_,J2_]:= I Deno[p,"<>ToString[InputForm[propmass]]<>","<>propsnames[[i]]<>"] SUNDelta[Gluonize[J1], Gluonize[J2]]"];
		propsrulesexp=Append[propsrulesexp,"I Deno[p,"<>ToString[InputForm[propmass]]<>","<>propsnames[[i]]<>"] SUNDelta[Gluonize[J1], Gluonize[J2]]"],
		propsrulestot=Append[propsrulestot,"prop" <> propsnames[[i]] <> "[m_,p_,J1_,J2_]:= I Deno[p,"<>ToString[InputForm[propmass]]<>","<>propsnames[[i]]<>"]"];
		propsrulesexp=Append[propsrulesexp,"I Deno[p,"<>ToString[InputForm[propmass]]<>","<>propsnames[[i]]<>"]"]]],
If[ToString[InputForm[propstypes[[i]]]]=="V",
LP7=LP6b/.myprops;
LP8=(LP7*(myc3*ME[-J2, -J3] + myc4*FV[p, -J2] FV[p, -J3]) - ME[-J1, -J3] // Expand) // MyContract;
LP9a = LP8 /. FV[x_, y_] -> 0;
LP9b = ((MySolve[LP9a, myc3] // MyContract) // Simplify);
LP10a = ((MySolve[LP8 /. LP9b, myc4] // MyContract) // Simplify);
If[ToString[InputForm[LP10a[[2]]]]=="ComplexInfinity",LP10a=myc4->0];
LP11=(-1)*(((myc3*ME[-J1, -J2] + myc4*FV[p, -J1] FV[p, -J2]) /. LP9b) /. LP10a);
LP12 = Parc[LP11 // Expand];
propfinaux=0;
Do[
propDenoaux="";
LPDeno = LP12[[j]] // Denominator;
LPNume = LP12[[j]] // Numerator;
If[ToString[InputForm[Level[LPDeno, 1][[1]]*Level[LPDeno, 1][[2]]]] == ToString[InputForm[LPDeno]],
Do[If[Coefficient[Level[LPDeno, 1][[k]],p,2]==-1,propmass=mysqrt[Level[LPDeno, 1][[k]]/.p->0],propmass=mysqrt[-Level[LPDeno, 1][[k]]/.p->0]];
propDenoaux=propDenoaux<>"Deno[p,"<>ToString[InputForm[propmass]]<>","<>propsnames[[i]]<>"]*";
propDenoaux2=ToExpression[StringDrop[propDenoaux,-1]],{k,1,Length[Level[LPDeno, 1]]}],
If[ToString[InputForm[Level[LPDeno, 1][[1]] + Level[LPDeno, 1][[2]]]] == ToString[InputForm[LPDeno]],
myfac=1;
If[Coefficient[LPDeno,p,2]==-1,myfac=-1;propmass=mysqrt[LPDeno/.p->0],propmass=mysqrt[-LPDeno/.p->0]];
propDenoaux2 = myfac*Deno[p,propmass,ToExpression[propsnames[[i]]]],
If[ToString[InputForm[Level[LPDeno, 1][[1]]^Level[LPDeno, 1][[2]]]] == ToString[InputForm[LPDeno]],
If[Level[LPDeno, 1][[2]]==2,propDenoaux2=Deno[p,0,ToExpression[propsnames[[i]]]],If[Level[LPDeno, 1][[2]]==4,propDenoaux2=Deno[p,0,ToExpression[propsnames[[i]]]]^2]]]]];
propfinaux += LPNume*propDenoaux2,{j,1,Length[LP12]}];
If[colcoin!=0&&ToString[InputForm[propsnames[[i]]]]==ToString[InputForm[myglu]],
	propsrulestot=Append[propsrulestot,"prop" <> propsnames[[i]] <> "[m_,p_,J1_,J2_]:= "<>ToString[InputForm[FullSimplify[propfinaux]]] <>" SUNDelta[Gluonize[J1], Gluonize[J2]]"];
	propsrulesexp=Append[propsrulesexp,ToString[InputForm[FullSimplify[propfinaux]]] <>" SUNDelta[Gluonize[J1], Gluonize[J2]]"],
	propsrulestot=Append[propsrulestot,"prop" <> propsnames[[i]] <> "[m_,p_,J1_,J2_]:= "<>ToString[InputForm[FullSimplify[propfinaux]]]];
	propsrulesexp=Append[propsrulesexp,ToString[InputForm[FullSimplify[propfinaux]]]]]]],{i,1,Length[propsnames]}],

(* Determine the rules for the propagators in a simplified version *)
Do[
propmass = ToString[ToExpression["m"<>palist2[[i,2]]] /. masstrade];
If[palist2[[i,1]]=="S" || palist2[[i,1]]=="U",
	If[colcoin!=0&&ToString[InputForm[palist2[[i,2]]]]==ToString[InputForm[myglugh]],
		propsrulestot=Append[propsrulestot,"prop" <> palist2[[i,2]] <> "[m_,p_,J1_,J2_]:= I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] SUNDelta[Gluonize[J1], Gluonize[J2]]"];
		propsrulesexp=Append[propsrulesexp,"I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] SUNDelta[Gluonize[J1], Gluonize[J2]]"],
		propsrulestot=Append[propsrulestot,"prop" <> palist2[[i,2]] <> "[m_,p_,J1_,J2_]:= I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"]"];
		propsrulesexp=Append[propsrulesexp,"I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"]"]],
If[palist2[[i,1]]=="V",
	If[colcoin!=0&&ToString[InputForm[palist2[[i,2]]]]==ToString[InputForm[myglu]],
		propsrulestot=Append[propsrulestot,"prop" <> palist2[[i,2]] <> "[m_,p_,J1_,J2_]:= -I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] ME[Index[Lorentz, Ext[1]], Index[Lorentz, Ext[2]]] SUNDelta[Gluonize[J1], Gluonize[J2]]"];
		propsrulesexp=Append[propsrulesexp,"-I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] ME[Index[Lorentz, Ext[1]], Index[Lorentz, Ext[2]]] SUNDelta[Gluonize[J1], Gluonize[J2]]"],
		propsrulestot=Append[propsrulestot,"prop" <> palist2[[i,2]] <> "[m_,p_,J1_,J2_]:= -I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] ME[Index[Lorentz, Ext[1]], Index[Lorentz, Ext[2]]]"];
		propsrulesexp=Append[propsrulesexp,"-I Deno[p,"<> propmass <>","<>palist2[[i,2]]<>"] ME[Index[Lorentz, Ext[1]], Index[Lorentz, Ext[2]]]"]]]],{i, 1, Length[palist2]}]
];

(* Determine the rules for the propagators in the unitary gauge *)
Do[
If[MemberQ[GaugeWithGold, palist3[[i, 2]]],
	propmass = ToString[ToExpression["m"<>palist2[[i,2]]] /. masstrade];
	propsrulestot=Append[propsrulestot,"prop" <> palist2[[i,2]] <> "UG[m_,p_,J1_,J2_]:= I Deno[p,"<> propmass <>","
	<>palist2[[i,2]]<>"] (-ME[Index[Lorentz, Ext[1]], Index[Lorentz, Ext[2]]] + fv[p,J1] fv[p,J2]/"<> propmass <>"^2)"]],{i, 1, Length[palist2]}];

(* Determine the rules for the fermionic propagators *)
Do[If[palist2[[i,1]]=="F",
propmass = ToString[ToExpression["m"<>palist2[[i,2]]] /. masstrade];
propsrulestot = Append[propsrulestot,"prop" <> palist2[[i,2]] <> "[m_,p_,J1_,J2_]:= I fprop[p," <> propmass <> "] Deno[p," <> propmass <> ","<>palist2[[i,2]]<>"]"];
propsrulesexp = Append[propsrulesexp,"I fprop[p," <> propmass <> "] Deno[p," <> propmass <> ","<>palist2[[i,2]]<>"]"];
propsnames = Append[propsnames,palist2[[i, 2]]]],{i, 1, Length[palist2]}];

(* * * * QGRAF INTERFACE * * * *)
fileQG = OpenWrite["built-model"];

Do[WriteString[fileQG, "[", auxprop[[i, 1]], ",", auxprop[[i, 2]], ",", auxprop[[i, 3]], "] \n"], {i, 1, Length[auxprop]}];
WriteString[fileQG, "\n \n"]

If[vHiggs != {},
	goodvHiggs = {};
	goodvHiggsStrList = {};
	Do[	If[ToString[InputForm[((vHiggs[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vHiggs[[i, 1]][[``, 1]], ",j]], {j,1,Length[vHiggs[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvHiggs, {" <> auxstr <> "}];"]], {i, 1, Length[vHiggs]}];
	Do[	auxstr="";
		Do[
			iaux=goodvHiggs[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvHiggs[[i]]]}];
		AppendTo[goodvHiggsStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvHiggs]}]];

If[vGauge != {},
	goodvGauge = {};
	goodvGaugeStrList = {};
	Do[	If[ToString[InputForm[((vGauge[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vGauge[[i, 1]][[``, 1]], ",j]], {j,1,Length[vGauge[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvGauge, {" <> auxstr <> "}];"]], {i, 1, Length[vGauge]}];
	Do[	auxstr="";
		Do[
			iaux=goodvGauge[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvGauge[[i]]]}];
		AppendTo[goodvGaugeStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvGauge]}]];

If[vFermions != {},
	goodvFermions = {};
	goodvFermionsStrList = {};
	Do[	If[ToString[InputForm[((vFermions[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vFermions[[i, 1]][[``, 1]], ",j]], {j,1,Length[vFermions[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvFermions, {" <> auxstr <> "}];"]], {i, 1, Length[vFermions]}];
	Do[	auxstr="";
		Do[
			iaux=goodvFermions[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvFermions[[i]]]}];
		AppendTo[goodvFermionsStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvFermions]}]];

If[vYukawa != {},
	goodvYukawa = {};
	goodvYukawaStrList = {};
	Do[	If[ToString[InputForm[((vYukawa[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vYukawa[[i, 1]][[``, 1]], ",j]], {j,1,Length[vYukawa[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvYukawa, {" <> auxstr <> "}];"]], {i, 1, Length[vYukawa]}];
	Do[	auxstr="";
		Do[
			iaux=goodvYukawa[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvYukawa[[i]]]}];
		AppendTo[goodvYukawaStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvYukawa]}]];

If[vGhosts != {},
	goodvGhosts = {};
	goodvGhostsStrList = {};
	Do[	If[ToString[InputForm[((vGhosts[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vGhosts[[i, 1]][[``, 1]], ",j]], {j,1,Length[vGhosts[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvGhosts, {" <> auxstr <> "}];"]], {i, 1, Length[vGhosts]}];
	Do[	auxstr="";
		Do[
			iaux=goodvGhosts[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvGhosts[[i]]]}];
		AppendTo[goodvGhostsStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvGhosts]}]];

If[vBeaks != {},
	goodvBeaks = {};
	goodvBeaksStrList = {};
	Do[	If[ToString[InputForm[((vBeaks[[i, 2]] // Simplify) /. MyRestr /. MyRestr)]] != "0", 
			auxstr="";
			Do[auxstr=auxstr <> ToString[StringForm["vBeaks[[i, 1]][[``, 1]], ",j]], {j,1,Length[vBeaks[[i,1]]]}];
			auxstr=StringDrop[auxstr,-2];
			ToExpression["AppendTo[goodvBeaks, {" <> auxstr <> "}];"]], {i, 1, Length[vBeaks]}];
	Do[	auxstr="";
		Do[
			iaux=goodvBeaks[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]] <> ",", {j,1,Length[goodvBeaks[[i]]]}];
		AppendTo[goodvBeaksStrList,StringDrop[auxstr,-1]];
		newstr="["<>StringDrop[auxstr,-1]<>"]\n";	
		WriteString[fileQG, newstr], {i, 1, Length[goodvBeaks]}]];

Close[fileQG];


(* * * * AMPERATE INTERFACE * * * *)
stmp = OpenWrite["PrePropagators.m"];
Do[WriteString[stmp, propsrulestot[[i]]<>"\n \n"], {i, 1, Length[propsrulestot]}];
Close[stmp];

stmp2 = OpenWrite["PrePrintPropagators.m"];
Do[WriteString[stmp2, "{{{{{{", propsnames[[i]],",",propsnames[[i]],"}}}}}},{{{{{{ ", propsrulesexp[[i]], " }}}}}}\n \n"], {i, 1, Length[propsrulesexp]}];
Close[stmp2];

vHiggsLogic = False;
If[vHiggs != {},
	vHiggsLogic = True;
	ExtvHiggs = {};
	Do[	myaux=((vHiggs[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvHiggs, ToString[InputForm[myaux]]]], {i, 1, Length[vHiggs]}];
	stmp = OpenWrite["PreRulesHiggs.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvHiggs[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvHiggs[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvHiggs[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvHiggs]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesHiggs.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvHiggsStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvHiggs[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvHiggs]}];
	Close[stmp2]];

vGaugeLogic = False;
If[vGauge != {},
	vGaugeLogic = True;
	ExtvGauge = {};
	Do[	myaux=((vGauge[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvGauge, ToString[InputForm[myaux]]]], {i, 1, Length[vGauge]}];
	stmp = OpenWrite["PreRulesGauge.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvGauge[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvGauge[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvGauge[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvGauge]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesGauge.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvGaugeStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvGauge[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvGauge]}];
	Close[stmp2]];

vFermionsLogic = False;
If[vFermions != {},
	vFermionsLogic = True;
	ExtvFermions = {};
	Do[	myaux=((vFermions[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvFermions, ToString[InputForm[myaux]]]], {i, 1, Length[vFermions]}];
	stmp = OpenWrite["PreRulesFermions.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvFermions[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvFermions[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvFermions[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvFermions]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesFermions.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvFermionsStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvFermions[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvFermions]}];
	Close[stmp2]];

vYukawaLogic = False;
If[vYukawa != {},
	vYukawaLogic = True;
	ExtvYukawa = {};
	Do[	myaux=((vYukawa[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvYukawa, ToString[InputForm[myaux]]]], {i, 1, Length[vYukawa]}];
	stmp = OpenWrite["PreRulesYukawa.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvYukawa[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvYukawa[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvYukawa[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvYukawa]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesYukawa.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvYukawaStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvYukawa[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvYukawa]}];
	Close[stmp2]];

vGhostsLogic = False;
If[vGhosts != {},
	vGhostsLogic = True;
	ExtvGhosts = {};
	Do[	myaux=((vGhosts[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvGhosts, ToString[InputForm[myaux]]]], {i, 1, Length[vGhosts]}];
	stmp = OpenWrite["PreRulesGhosts.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvGhosts[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvGhosts[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvGhosts[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvGhosts]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesGhosts.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvGhostsStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvGhosts[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvGhosts]}];
	Close[stmp2]];

vBeaksLogic = False;
If[vBeaks != {},
	vBeaksLogic = True;
	ExtvBeaks = {};
	Do[	myaux=((vBeaks[[i, 2]] // Simplify) //. MyRestr);
		If[ToString[InputForm[myaux]]!="0", AppendTo[ExtvBeaks, ToString[InputForm[myaux]]]], {i, 1, Length[vBeaks]}];
	stmp = OpenWrite["PreRulesBeaks.m"];
	Do[	auxstr="vrtx";
		ancstr="[";
		Do[
			iaux=goodvBeaks[[i, j]];
			auxstr=auxstr <> ToString[InputForm[iaux]];
			ancstr=ancstr<>"J"<>ToString[j]<>"_,"<>"k"<>ToString[j]<>"_,", {j,1,Length[goodvBeaks[[i]]]}];
		ancstr=StringDrop[ancstr,-1]<>"]:= ";
		newstr=auxstr<>ancstr<>ExtvBeaks[[i]]<>" \n \n";
		WriteString[stmp,newstr],{i, 1, Length[goodvBeaks]}];
	Close[stmp];
	stmp2 = OpenWrite["PrePrintRulesBeaks.m"];
	Do[WriteString[stmp2, "{{{{{{" <> goodvBeaksStrList[[i]] <> "}}}}}},{{{{{{ " <> ExtvBeaks[[i]] <> " }}}}}}\n \n"], {i, 1, Length[goodvBeaks]}];
	Close[stmp2]];

(* Writing the Feynman Rules main file *)
str = OpenWrite["Feynman-Rules-Main.m"];
WriteString[str, "(* ----------- Some Extras ----------- *) \n"];
(* Loading Extras like DecayWidths and Neutrinos list *)
WriteString[str, "Get[\"Extras.m\", Path -> {dirFey}]\n"];
WriteString[str, "\n"];
(* Deno is non-trivial should there be decay widths *)
WriteString[str, "(* ----------- Deno Function ----------- *) \n"];
WriteString[str, "Deno = Function[{k, m, PP}, Module[{mycoin, myres, decay13check, scat23check},\n"];
WriteString[str, "If[loops==0 && Length[inparticlesB] == 1 && Length[outparticlesB] == 3,decay13check=1,decay13check=0];\n"];
WriteString[str, "If[loops==0 && Length[inparticlesB] == 2 && Length[outparticlesB] == 3,scat23check=1,scat23check=0];\n"];
WriteString[str, "If[ToString[InputForm[k]] != \"-p1 - p2\" && decay13check==0 && scat23check==0,\n"];
WriteString[str, "	myres = FAD[{k, m}, Dimension -> D],\n"];
WriteString[str, "	mycoin = 0;\n"];
WriteString[str, "	Do[If[ToString[PP] == ToString[DWs[[i, 1]]],\n"];
WriteString[str, "			mycoin += 1;\n"];
WriteString[str, " 	 		myres = (SPD[k, k] - m^2 + I*m*DWs[[i, 2]])^(-1)], {i, 1, Length[DWs]}];\n"];
WriteString[str, "If[mycoin == 0, myres = FAD[{k, m}, Dimension -> D]]];\n"];
WriteString[str, "myres]];\n\n"];
(* We go on to some definitions *)
WriteString[str, "(* ----------- Definitions ----------- *)\n"];
WriteString[str, "dm[mu_]:=DiracMatrix[mu,Dimension->D] \n"];
WriteString[str, "dm[5]:=DiracMatrix[5] \n"];
WriteString[str, "ds[p_]:=DiracSlash[p, Dimension -> D] \n"];
WriteString[str, "fprop[p_,m_]:=ds[p] + m \n"];
WriteString[str, "fv[p_,mu_]:=FourVector[p,mu,Dimension->D] \n"];
WriteString[str, "po2[p_]:=Pair[Momentum[p], Momentum[p]] \n"];
WriteString[str, "mt[mu_,nu_]:=MetricTensor[mu,nu,Dimension->D] \n"];
WriteString[str, "sp[p_,q_]:=ScalarProduct[p,q,Dimension->D] \n \n"];
WriteString[str, "(* ----------- Gluonize and mtom: useful functions for gluon indices ----------- *)\n"];
WriteString[str, "Gluonize = Function[exp, Module[{myres},\n"];
WriteString[str, "   If[StringTake[ToString[exp], 1] == \"-\", \n"];
WriteString[str, "      myres = ToExpression[\"-G\" <> StringDrop[ToString[exp], 2]],\n"];
WriteString[str, "      myres = ToExpression[\"G\" <> StringDrop[ToString[exp], 1]]];\n"];
WriteString[str, "      myres]];\n"];
WriteString[str, "mtom = Function[exp, Module[{myres},\n"];
WriteString[str, "   myres = ToExpression[StringReplace[exp, \"-\" -> \"m\"]];\n"];
WriteString[str, "   myres]];\n\n"];
WriteString[str, "(* ----------- Mass conversions ----------- *)\n"];
Do[If[ToString[masstrade[[i,1]]]!=ToString[masstrade[[i,2]]],
WriteString[str, ToString[masstrade[[i,1]]]<>":="<>ToString[masstrade[[i,2]]]<>"\n"]],{i,1,Length[masstrade]}];
WriteString[str, "\n"];
WriteString[str, "(* ----------- Set Tadpole momentum to zero ----------- *)\n"];
WriteString[str, "J0:=0 \n"];
WriteString[str, "\n"];
WriteString[str, "(* ----------- Feynman Rules ----------- *) \n"];
(* Loading the different Feynman rules files in the main file*)
WriteString[str, "Get[\"Propagators.m\", Path -> {dirFey}]\n"];
If[vHiggsLogic,WriteString[str, "Get[\"RulesHiggs.m\", Path -> {dirFey}]\n"]];
If[vGaugeLogic,WriteString[str, "Get[\"RulesGauge.m\", Path -> {dirFey}]\n"]];
If[vFermionsLogic,WriteString[str, "Get[\"RulesFermions.m\", Path -> {dirFey}]\n"]];
If[vYukawaLogic,WriteString[str, "Get[\"RulesYukawa.m\", Path -> {dirFey}]\n"]];
If[vGhostsLogic,WriteString[str, "Get[\"RulesGhosts.m\", Path -> {dirFey}]\n"]];
If[vBeaksLogic,WriteString[str, "Get[\"RulesBeaks.m\", Path -> {dirFey}]\n"]];
WriteString[str, "\n"];
Close[str];


(* Protecting variables in use: *)
If[!RenoLogic,
ClearAll[variableQ];
variableQ[name_String] := MatchQ[ToExpression[name, InputForm, OwnValues], 
  					Except[{} | {_ :> (Function | CompiledFunction)[___]}]];
ToProtect = Select[Names["Global`*"], variableQ];
Do[ToString[StringForm[ToProtect[[i]] <> "//Protect"]] // ToExpression,{i,1,Length[ToProtect]}]];
